@@ -9,11 +9,11 @@ from pathlib import Path
 from subprocess import CalledProcessError
 from tempfile import TemporaryDirectory
 from ubuntu_image.helpers import (
-     DoesNotFit, MiB, mkfs_ext4, run)
+     DoesNotFit, MiB, mkfs_ext4, run, unsparse_swapfile_ext4)
 from ubuntu_image.hooks import HookManager
 from ubuntu_image.image import Image
 from ubuntu_image.parser import (
-    FileSystemType, StructureRole, VolumeSchema,
+    BootLoader, FileSystemType, StructureRole, VolumeSchema,
     parse as parse_yaml)
 from ubuntu_image.state import State
 
@@ -59,6 +59,8 @@ class AbstractImageBuilderState(State):
         self.unpackdir = None
         self.volumedir = None
         self.cloud_init = args.cloud_init
+        self.disk_info = args.disk_info
+        self.disable_console_conf = args.disable_console_conf
         self.exitcode = 0
         self.done = False
         # Generic hook handling manager.
@@ -71,6 +73,8 @@ class AbstractImageBuilderState(State):
         state.update(
             args=self.args,
             cloud_init=self.cloud_init,
+            disk_info=self.disk_info,
+            disable_console_conf=self.disable_console_conf,
             done=self.done,
             exitcode=self.exitcode,
             gadget=self.gadget,
@@ -90,6 +94,8 @@ class AbstractImageBuilderState(State):
         super().__setstate__(state)
         self.args = state['args']
         self.cloud_init = state['cloud_init']
+        self.disk_info = state['disk_info']
+        self.disable_console_conf = state['disable_console_conf']
         self.done = state['done']
         self.exitcode = state['exitcode']
         self.gadget = state['gadget']
@@ -152,8 +158,23 @@ class AbstractImageBuilderState(State):
 
     def populate_rootfs_contents_hooks(self):
         # Separate populate step for firing the post-populate-rootfs hook.
-        env = {'UBUNTU_IMAGE_HOOK_ROOTFS': self.rootfs}
-        self.hook_manager.fire('post-populate-rootfs', env)
+        if self.gadget.seeded:
+            # Running this hook for model assertion builds is a bit weird,
+            # but makes even less sense for UC20 models.  Disable.
+            _logger.debug('Building from a seeded gadget - skipping the '
+                          'post-populate-rootfs hook execution: unsupported.')
+        else:
+            env = {'UBUNTU_IMAGE_HOOK_ROOTFS': self.rootfs}
+            self.hook_manager.fire('post-populate-rootfs', env)
+        self._next.append(self.generate_disk_info)
+
+    def generate_disk_info(self):
+        # If available, populate the rootfs .disk/info
+        if self.disk_info is not None:
+            disk_info_dir = os.path.join(self.rootfs, '.disk')
+            os.makedirs(disk_info_dir, exist_ok=True)
+            disk_info_file = os.path.join(disk_info_dir, 'info')
+            shutil.copy(self.disk_info, disk_info_file)
         self._next.append(self.calculate_rootfs_size)
 
     @staticmethod
@@ -190,18 +211,128 @@ class AbstractImageBuilderState(State):
                 os.makedirs(target_dir, exist_ok=True)
         self._next.append(self.populate_bootfs_contents)
 
+    @classmethod
+    def _selective_copytree(cls, src, dst):
+        # Selectively copy entries from src to dst directory. Entries already
+        # existing at dst overwritten.
+        os.makedirs(dst, exist_ok=True)
+        shutil.copystat(src, dst)
+        for entry in os.scandir(src):
+            sname = os.path.join(src, entry.name)
+            dname = os.path.join(dst, entry.name)
+            if entry.is_file() or entry.is_symlink():
+                if os.path.exists(dname):
+                    continue
+                # copy files and recreate symlinks
+                shutil.copy2(sname, dname, follow_symlinks=False)
+            elif entry.is_dir():
+                # recursive copy directories
+                cls._selective_copytree(sname, dname)
+            else:
+                # will most likely raise shutil.SpecialFileError
+                shutil.copy2(sname, dname, follow_symlinks=False)
+
+    def _should_skip_partition(self, part):
+        # Helper used to determine if a partition should be acted on or not.
+        return self.gadget.seeded and part.role in (
+            StructureRole.system_boot,
+            StructureRole.system_data,
+            StructureRole.system_save)
+
+    def _populate_one_bootfs(self, name, volume):
+        for partnum, part in enumerate(volume.structures):
+            if self._should_skip_partition(part):
+                continue
+            if part.role is StructureRole.system_seed:
+                # For seeded systems, the system-seed partition (which reuses
+                # the rootfs paths) is also the boot partition - so we need
+                # to redirect all the boot copies there as well.
+                target_dir = self.rootfs
+            else:
+                target_dir = os.path.join(
+                    volume.basedir, 'part{}'.format(partnum))
+            if part.role in (StructureRole.system_boot,
+                             StructureRole.system_seed):
+                volume.bootfs = target_dir
+                if volume.bootloader is BootLoader.uboot:
+                    boot = os.path.join(
+                        self.unpackdir, 'image', 'boot', 'uboot')
+                    ubuntu = target_dir
+                elif volume.bootloader is BootLoader.grub:
+                    boot = os.path.join(
+                        self.unpackdir, 'image', 'boot', 'grub')
+                    # XXX: Bad special-casing.  `snap prepare-image` currently
+                    # installs to /boot/grub, but we need to map this to
+                    # /EFI/ubuntu.  This is because we are using a SecureBoot
+                    # signed bootloader image which has this path embedded, so
+                    # we need to install our files to there.
+                    ubuntu = os.path.join(target_dir, 'EFI', 'ubuntu')
+                else:
+                    raise ValueError(
+                        'Unsupported volume bootloader value: {}'.format(
+                            volume.bootloader))
+                if os.path.isdir(boot):
+                    os.makedirs(ubuntu, exist_ok=True)
+                    for filename in os.listdir(boot):
+                        src = os.path.join(boot, filename)
+                        dst = os.path.join(ubuntu, filename)
+                        # XXX: Use _selective_copytree?
+                        shutil.move(src, dst)
+                else:
+                    _logger.debug('No bootloader bits prepared in the rootfs '
+                                  '- skipping boot copies.')
+            gadget_dir = os.path.join(self.unpackdir, 'gadget')
+            if part.filesystem is not FileSystemType.none:
+                for content in part.content:
+                    src = os.path.join(gadget_dir, content.source)
+                    dst = os.path.join(target_dir, content.target)
+                    if content.source.endswith('/'):
+                        # This is a directory copy specification.  The target
+                        # must also end in a slash.
+                        #
+                        # XXX: If this is a file instead of a directory, give
+                        # a useful error message instead of a traceback.
+                        #
+                        # XXX: We should assert this constraint in the parser.
+                        target, slash, tail = content.target.rpartition('/')
+                        if slash != '/' and tail != '':
+                            raise ValueError(
+                                'target must end in a slash: {}'.format(
+                                    content.target))
+                        # The target of a recursive directory copy is the
+                        # target directory name, with or without a trailing
+                        # slash necessary at least to handle the case of
+                        # recursive copy into the root directory), so make
+                        # sure here that it exists.
+                        os.makedirs(dst, exist_ok=True)
+                        for filename in os.listdir(src):
+                            sub_src = os.path.join(src, filename)
+                            dst = os.path.join(target_dir, target, filename)
+                            if os.path.isdir(sub_src):
+                                self._selective_copytree(sub_src, dst)
+                            else:
+                                if not os.path.exists(dst):
+                                    shutil.copy(sub_src, dst)
+                    else:
+                        # XXX: If this is a directory instead of a file, give
+                        # a useful error message instead of a traceback.
+                        os.makedirs(os.path.dirname(dst), exist_ok=True)
+                        if not os.path.exists(dst):
+                            shutil.copy(src, dst)
+
     def populate_bootfs_contents(self):
-        # Abstract, should be re-implemented by derivatives.
+        for name, volume in self.gadget.volumes.items():
+            self._populate_one_bootfs(name, volume)
         self._next.append(self.prepare_filesystems)
 
     def _prepare_one_volume(self, volume_index, name, volume):
         volume.part_images = []
         farthest_offset = 0
         for partnum, part in enumerate(volume.structures):
-            part_img = os.path.join(
-                volume.basedir, 'part{}.img'.format(partnum))
-            if part.role is StructureRole.system_data:
-                # The image for the root partition.
+            # The system-data and system-seed partitions do not have to have
+            # an explicit size set.
+            if part.role in (StructureRole.system_data,
+                             StructureRole.system_seed):
                 if part.size is None:
                     part.size = self.rootfs_size
                 elif part.size < self.rootfs_size:
@@ -209,6 +340,16 @@ class AbstractImageBuilderState(State):
                                     'actual rootfs contents {}'.format(
                                         part.size, self.rootfs_size))
                     part.size = self.rootfs_size
+            # We do the set for farthest offset now as for image sizing we
+            # still need to consider partitions that we 'skip'.
+            farthest_offset = max(farthest_offset, (part.offset + part.size))
+            if self._should_skip_partition(part):
+                continue
+            part_img = os.path.join(
+                volume.basedir, 'part{}.img'.format(partnum))
+            # Create the actual image files now.
+            if part.role is StructureRole.system_data:
+                # The image for the root partition.
                 # We defer creating the root file system image because we have
                 # to populate it at the same time.  See mkfs.ext4(8) for
                 # details.
@@ -228,7 +369,6 @@ class AbstractImageBuilderState(State):
                     run('mkfs.vfat -s 1 -S 512 -F 32 {} {}'.format(
                         label_option, part_img))
             volume.part_images.append(part_img)
-            farthest_offset = max(farthest_offset, (part.offset + part.size))
         # Calculate or check the final image size.
         #
         # TODO: Hard-codes last 34 512-byte sectors for backup GPT,
@@ -279,15 +419,43 @@ class AbstractImageBuilderState(State):
         self._next.append(self.populate_filesystems)
 
     def _populate_one_volume(self, name, volume):
+        # For the LK bootloader we need to copy boot.img and snapbootsel.bin to
+        # the gadget folder so they can be used as partition content. The first
+        # one comes from the kernel snap, while the second one is modified by
+        # 'snap prepare-image' to set the right core and kernel for the kernel
+        # command line.
+        if volume.bootloader is BootLoader.lk:
+            boot = os.path.join(
+                self.unpackdir, 'image', 'boot', 'lk')
+            gadget = os.path.join(
+                self.unpackdir, 'gadget')
+            if os.path.isdir(boot):
+                os.makedirs(gadget, exist_ok=True)
+                for filename in os.listdir(boot):
+                    src = os.path.join(boot, filename)
+                    dst = os.path.join(gadget, filename)
+                    shutil.copy(src, dst)
         for partnum, part in enumerate(volume.structures):
+            if self._should_skip_partition(part):
+                continue
             part_img = volume.part_images[partnum]
-            part_dir = os.path.join(volume.basedir, 'part{}'.format(partnum))
+            # In seeded images, the system-seed partition is basically the
+            # rootfs partition - at least from the ubuntu-image POV.
+            if part.role is StructureRole.system_seed:
+                part_dir = self.rootfs
+            else:
+                part_dir = os.path.join(volume.basedir,
+                                        'part{}'.format(partnum))
             if part.role is StructureRole.system_data:
                 # The root partition needs to be ext4, which may or may not be
                 # populated at creation time, depending on the version of
                 # e2fsprogs.
-                mkfs_ext4(part_img, self.rootfs, part.filesystem_label,
-                          preserve_ownership=True)
+                mkfs_ext4(part_img, self.rootfs, self.args.cmd,
+                          part.filesystem_label, preserve_ownership=True)
+                # XXX: This is a workaround for mkfs.ext4 sparsifying our
+                # classic rootfses swapfile.
+                if self.args.cmd == 'classic':
+                    unsparse_swapfile_ext4(part_img)
             elif part.filesystem is FileSystemType.none:
                 image = Image(part_img, part.size)
                 offset = 0
@@ -327,7 +495,8 @@ class AbstractImageBuilderState(State):
                 run('mcopy -s -i {} {} ::'.format(part_img, sourcefiles),
                     env=env)
             elif part.filesystem is FileSystemType.ext4:
-                mkfs_ext4(part_img, part_dir, part.filesystem_label)
+                mkfs_ext4(part_img, part_dir, self.args.cmd,
+                          part.filesystem_label)
             else:
                 raise AssertionError('Invalid part filesystem type: {}'.format(
                     part.filesystem))
@@ -343,13 +512,18 @@ class AbstractImageBuilderState(State):
         image = Image(imgfile, volume.image_size, volume.schema)
         offset_writes = []
         part_offsets = {}
-        # We first create all the partitions.
+        # We first create all the needed partitions.
+        # For regular core16 and core18 builds, this means creating all of the
+        # defined partitions.  For core20 (the so called 'seeded images'), we
+        # only create all the role-less partitions and mbr + system-seed.
+        # The rest is created dynamically by snapd on first boot.
         for part in volume.structures:
             if part.name is not None:
                 part_offsets[part.name] = part.offset
             if part.offset_write is not None:
                 offset_writes.append((part.offset, part.offset_write))
-            if part.role is StructureRole.mbr or part.type == 'bare':
+            if (part.role is StructureRole.mbr or part.type == 'bare' or
+                    self._should_skip_partition(part)):
                 continue
             activate = False
             if (volume.schema is VolumeSchema.mbr and
@@ -366,6 +540,8 @@ class AbstractImageBuilderState(State):
         # clobbers things like hybrid MBR partitions.
         part_id = 1
         for i, part in enumerate(volume.structures):
+            if self._should_skip_partition(part):
+                continue
             image.copy_blob(volume.part_images[i],
                             bs=image.sector_size,
                             seek=part.offset // image.sector_size,
