@@ -25,6 +25,7 @@ __all__ = [
     'run',
     'snap',
     'sparse_copy',
+    'unsparse_swapfile_ext4',
     ]
 
 
@@ -65,7 +66,7 @@ def straight_up_bytes(count):
 
 
 def as_size(size, min=0, max=None):
-    mo = re.match('(\d+)([a-zA-Z]*)', size)
+    mo = re.match(r'(\d+)([a-zA-Z]*)', size)
     if mo is None:
         raise ValueError(size)
     size_in_bytes = mo.group(1)
@@ -134,7 +135,7 @@ def snap(model_assertion, root_dir, channel=None, extra_snaps=None):
         arg_list.append('--channel={}'.format(channel))
     # Fails if extra_snaps is None or the empty list.
     if extra_snaps:
-        arg_list.append(SPACE.join('--extra-snaps={}'.format(extra)
+        arg_list.append(SPACE.join('--snap={}'.format(extra)
                         for extra in extra_snaps))
     arg_list.extend([model_assertion, root_dir])
     cmd = SPACE.join(arg_list)
@@ -209,7 +210,18 @@ def mount(img):
         yield mountpoint
 
 
-def mkfs_ext4(img_file, contents_dir, label='writable',
+def unsparse_swapfile_ext4(img_file):
+    # A workaround for mkfs.ext4 messing up our rootfs swapfiles by turning
+    # them to sparse files without us asking.
+    with mount(img_file) as mountpoint:
+        swapfile_path = os.path.join(mountpoint, 'swapfile')
+        if os.path.exists(swapfile_path):
+            cmd = ('dd if={swapfile} of={swapfile} conv=notrunc '
+                   'bs=1M'.format(swapfile=swapfile_path))
+            run(cmd, stdout=None, stderr=None, capture_output=False)
+
+
+def mkfs_ext4(img_file, contents_dir, image_type, label='writable',
               preserve_ownership=False):
     """Encapsulate the `mkfs.ext4` invocation.
 
@@ -218,10 +230,21 @@ def mkfs_ext4(img_file, contents_dir, label='writable',
     contents of an existing directory.  Unfortunately, we're targeting
     Ubuntu 16.04, which has e2fsprogs 1.42.X without the -d flag.  In
     that case, we have to sudo loop mount the ext4 file system and
-    populate it that way.  Which sucks because sudo.
+    populate it that way.
+
+    Besides that, we need to use fakeroot for core so the files in the
+    partition are owned by root. For classic we want to keep the
+    ownership of the files created by live-build, so we do not use
+    fakeroot. But, we need to use sudo in that case so we can access
+    root read-only files in the contents folder.
     """
-    cmd = ('fakeroot-sysv mkfs.ext4 -L {} -O -metadata_csum -T default '
-           '-O uninit_bg {} -d {}').format(label, img_file, contents_dir)
+    if image_type == 'snap':
+        sudo_cmd = 'fakeroot-sysv'
+    else:
+        sudo_cmd = 'sudo'
+    cmd = ('{} mkfs.ext4 -L {} -O -metadata_csum -T default '
+           '-O uninit_bg {} -d {}').format(sudo_cmd, label, img_file,
+                                           contents_dir)
     proc = run(cmd, check=False)
     if proc.returncode == 0:
         # We have a new enough e2fsprogs, so we're done.
