@@ -13,9 +13,10 @@ from subprocess import run as subprocess_run
 from tempfile import NamedTemporaryFile, TemporaryDirectory
 from types import SimpleNamespace
 from ubuntu_image.helpers import (
-     GiB, MiB, PrivilegeError, as_bool, as_size,
-     check_root_privilege, get_host_arch, get_host_distro, live_build,
-     mkfs_ext4, run, snap, sparse_copy)
+     DependencyError, GiB, MiB, PrivilegeError, as_bool, as_size,
+     check_root_privilege, get_host_arch, get_host_distro,
+     get_qemu_static_for_arch, live_build,
+     mkfs_ext4, run, snap, sparse_copy, unsparse_swapfile_ext4)
 from ubuntu_image.testing.helpers import (
      LiveBuildMocker, LogCapture, envar)
 from unittest import TestCase
@@ -64,10 +65,12 @@ def is_sparse(path):
 
 
 class MountMocker:
-    def __init__(self, results_dir):
+    def __init__(self, results_dir, contents_dir=None):
         self.mountpoint = None
         self.results_dir = results_dir
         self.preserves_ownership = False
+        self.contents_dir = contents_dir
+        self.dd_called = False
 
     def run(self, command, *args, **kws):
         if 'mkfs.ext4' in command:
@@ -83,6 +86,10 @@ class MountMocker:
             # which will be a temporary directory, so that we can verify its
             # contents later.
             self.mountpoint = command.split()[-1]
+            if self.contents_dir:
+                subprocess_run('cp {}/* {}'.format(
+                    self.contents_dir, self.mountpoint), shell=True,
+                    capture_output=False)
         elif command.startswith('sudo umount'):
             # Just ignore the umount command since we never mounted anything,
             # and it's a temporary directory anyway.
@@ -100,6 +107,10 @@ class MountMocker:
             # way of mocking this as everything else would require root.
             if re.search(r'--preserve=[^ ]*ownership', command):
                 self.preserves_ownership = True
+        elif command.startswith('dd'):
+            # dd is a safe command so we should just run it.
+            subprocess_run(command, shell=True, capture_output=False)
+            self.dd_called = True
 
 
 class TestHelpers(TestCase):
@@ -174,6 +185,17 @@ class TestHelpers(TestCase):
     def test_get_host_distro(self):
         self.assertIsNotNone(get_host_distro())
 
+    def test_get_qemu_static_for_arch(self):
+        executables = {
+            'armhf': 'qemu-arm-static',
+            'arm64': 'qemu-aarch64-static',
+            'ppc64el': 'qemu-ppc64le-static',
+            'i386': 'qemu-i386-static',
+            'foo': 'qemu-foo-static',
+            }
+        for arch, static in executables.items():
+            self.assertEqual(get_qemu_static_for_arch(arch), static)
+
     def test_sparse_copy(self):
         with ExitStack() as resources:
             tmpdir = resources.enter_context(TemporaryDirectory())
@@ -238,12 +260,13 @@ class TestHelpers(TestCase):
                 patch('ubuntu_image.helpers.subprocess_run',
                       return_value=FakeProc()))
             tmpdir = resources.enter_context(TemporaryDirectory())
-            snap(model, tmpdir, extra_snaps=('foo', 'bar'))
+            snap(model, tmpdir, extra_snaps=('foo', 'bar=edge', 'baz=18/beta'))
             self.assertEqual(len(mock.call_args_list), 1)
             args, kws = mock.call_args_list[0]
         self.assertEqual(
             args[0],
-            ['snap', 'prepare-image', '--extra-snaps=foo', '--extra-snaps=bar',
+            ['snap', 'prepare-image',
+             '--snap=foo', '--snap=bar=edge', '--snap=baz=18/beta',
              model, tmpdir])
 
     def test_live_build(self):
@@ -254,6 +277,9 @@ class TestHelpers(TestCase):
             resources.enter_context(LogCapture())
             resources.enter_context(
                 patch('ubuntu_image.helpers.run', mock.run))
+            resources.enter_context(
+                patch('ubuntu_image.helpers.get_host_arch',
+                      return_value='amd64'))
             env = OrderedDict()
             env['PROJECT'] = 'ubuntu-server'
             env['SUITE'] = 'xenial'
@@ -279,6 +305,9 @@ class TestHelpers(TestCase):
             resources.enter_context(LogCapture())
             resources.enter_context(
                 patch('ubuntu_image.helpers.run', mock.run))
+            resources.enter_context(
+                patch('ubuntu_image.helpers.get_host_arch',
+                      return_value='amd64'))
             env = OrderedDict()
             env['PROJECT'] = 'ubuntu-cpc'
             env['SUITE'] = 'xenial'
@@ -317,6 +346,9 @@ class TestHelpers(TestCase):
             resources.enter_context(LogCapture())
             resources.enter_context(
                 patch('ubuntu_image.helpers.run', mock.run))
+            resources.enter_context(
+                patch('ubuntu_image.helpers.get_host_arch',
+                      return_value='amd64'))
             resources.enter_context(
                 envar('UBUNTU_IMAGE_LIVECD_ROOTFS_AUTO_PATH', auto_dir))
             env = OrderedDict()
@@ -341,7 +373,130 @@ class TestHelpers(TestCase):
             with open(os.path.join(config_path), 'r') as fp:
                 self.assertEqual(fp.read(), 'DUMMY')
 
-    def test_mkfs_ext4(self):
+    def test_live_build_cross_build(self):
+        with ExitStack() as resources:
+            tmpdir = resources.enter_context(TemporaryDirectory())
+            root_dir = os.path.join(tmpdir, 'root_dir')
+            mock = LiveBuildMocker(root_dir)
+            resources.enter_context(LogCapture())
+            resources.enter_context(
+                patch('ubuntu_image.helpers.run', mock.run))
+            resources.enter_context(
+                patch('ubuntu_image.helpers.get_host_arch',
+                      return_value='amd64'))
+            resources.enter_context(
+                patch('ubuntu_image.helpers.find_executable',
+                      return_value='/usr/bin/qemu-arm-static-fake'))
+            env = OrderedDict()
+            env['PROJECT'] = 'ubuntu-server'
+            env['SUITE'] = 'xenial'
+            env['ARCH'] = 'armhf'
+            live_build(root_dir, env)
+            self.assertEqual(len(mock.call_args_list), 3)
+            self.assertEqual(
+                mock.call_args_list[1],
+                ['sudo',
+                 'PROJECT=ubuntu-server', 'SUITE=xenial', 'ARCH=armhf',
+                 'lb', 'config',
+                 '--bootstrap-qemu-arch', 'armhf',
+                 '--bootstrap-qemu-static', '/usr/bin/qemu-arm-static-fake',
+                 '--architectures', 'armhf'])
+            self.assertEqual(
+                mock.call_args_list[2],
+                ['sudo',
+                 'PROJECT=ubuntu-server', 'SUITE=xenial', 'ARCH=armhf',
+                 'lb', 'build'])
+
+    def test_live_build_cross_build_env_path(self):
+        with ExitStack() as resources:
+            tmpdir = resources.enter_context(TemporaryDirectory())
+            root_dir = os.path.join(tmpdir, 'root_dir')
+            mock = LiveBuildMocker(root_dir)
+            resources.enter_context(LogCapture())
+            resources.enter_context(
+                patch('ubuntu_image.helpers.run', mock.run))
+            resources.enter_context(
+                patch('ubuntu_image.helpers.get_host_arch',
+                      return_value='amd64'))
+            resources.enter_context(
+                patch('ubuntu_image.helpers.find_executable',
+                      return_value='/usr/bin/qemu-arm-static'))
+            resources.enter_context(
+                envar('UBUNTU_IMAGE_QEMU_USER_STATIC_PATH',
+                      '/opt/qemu-arm-static'))
+            env = OrderedDict()
+            env['PROJECT'] = 'ubuntu-server'
+            env['SUITE'] = 'xenial'
+            env['ARCH'] = 'armhf'
+            live_build(root_dir, env)
+            self.assertEqual(len(mock.call_args_list), 3)
+            self.assertEqual(
+                mock.call_args_list[1],
+                ['sudo',
+                 'PROJECT=ubuntu-server', 'SUITE=xenial', 'ARCH=armhf',
+                 'lb', 'config',
+                 '--bootstrap-qemu-arch', 'armhf',
+                 '--bootstrap-qemu-static', '/opt/qemu-arm-static',
+                 '--architectures', 'armhf'])
+            self.assertEqual(
+                mock.call_args_list[2],
+                ['sudo',
+                 'PROJECT=ubuntu-server', 'SUITE=xenial', 'ARCH=armhf',
+                 'lb', 'build'])
+
+    def test_live_build_cross_build_no_static(self):
+        with ExitStack() as resources:
+            tmpdir = resources.enter_context(TemporaryDirectory())
+            root_dir = os.path.join(tmpdir, 'root_dir')
+            mock = LiveBuildMocker(root_dir)
+            resources.enter_context(LogCapture())
+            resources.enter_context(
+                patch('ubuntu_image.helpers.run', mock.run))
+            resources.enter_context(
+                patch('ubuntu_image.helpers.get_host_arch',
+                      return_value='amd64'))
+            resources.enter_context(
+                patch('ubuntu_image.helpers.find_executable',
+                      return_value=None))
+            env = OrderedDict()
+            env['PROJECT'] = 'ubuntu-server'
+            env['SUITE'] = 'xenial'
+            env['ARCH'] = 'armhf'
+            with self.assertRaises(DependencyError) as cm:
+                live_build(root_dir, env)
+            self.assertEqual(len(mock.call_args_list), 1)
+            self.assertEqual(
+                cm.exception.name, 'qemu-arm-static')
+
+    def test_live_build_no_cross_build(self):
+        with ExitStack() as resources:
+            tmpdir = resources.enter_context(TemporaryDirectory())
+            root_dir = os.path.join(tmpdir, 'root_dir')
+            mock = LiveBuildMocker(root_dir)
+            resources.enter_context(LogCapture())
+            resources.enter_context(
+                patch('ubuntu_image.helpers.run', mock.run))
+            resources.enter_context(
+                patch('ubuntu_image.helpers.get_host_arch',
+                      return_value='amd64'))
+            resources.enter_context(
+                patch('ubuntu_image.helpers.find_executable',
+                      return_value='/usr/bin/qemu-arm-static'))
+            env = OrderedDict()
+            env['PROJECT'] = 'ubuntu-server'
+            env['SUITE'] = 'xenial'
+            env['ARCH'] = 'armhf'
+            live_build(root_dir, env, enable_cross_build=False)
+            # Make sure that if we explicity disable cross-building, no
+            # cross-build arguments are passed to lb config
+            self.assertEqual(len(mock.call_args_list), 3)
+            self.assertEqual(
+                mock.call_args_list[1],
+                ['sudo',
+                 'PROJECT=ubuntu-server', 'SUITE=xenial', 'ARCH=armhf',
+                 'lb', 'config'])
+
+    def aux_test_mkfs_ext4(self, cmd):
         with ExitStack() as resources:
             tmpdir = resources.enter_context(TemporaryDirectory())
             results_dir = os.path.join(tmpdir, 'results')
@@ -356,7 +511,7 @@ class TestHelpers(TestCase):
                 fp.write(b'56789')
             # And a fake image file.
             img_file = resources.enter_context(NamedTemporaryFile())
-            mkfs_ext4(img_file, contents_dir)
+            mkfs_ext4(img_file, contents_dir, cmd)
             # Two files were put in the "mountpoint" directory, but because of
             # above, we have to check them in the results copy.
             with open(os.path.join(mock.results_dir, 'a.dat'), 'rb') as fp:
@@ -364,6 +519,12 @@ class TestHelpers(TestCase):
             with open(os.path.join(mock.results_dir, 'b.dat'), 'rb') as fp:
                 self.assertEqual(fp.read(), b'56789')
 
+    def test_mkfs_ext4_snap(self):
+        self.aux_test_mkfs_ext4('snap')
+
+    def test_mkfs_ext4_classic(self):
+        self.aux_test_mkfs_ext4('classic')
+
     def test_mkfs_ext4_no_contents(self):
         with ExitStack() as resources:
             tmpdir = resources.enter_context(TemporaryDirectory())
@@ -375,7 +536,7 @@ class TestHelpers(TestCase):
             contents_dir = resources.enter_context(TemporaryDirectory())
             # And a fake image file.
             img_file = resources.enter_context(NamedTemporaryFile())
-            mkfs_ext4(img_file, contents_dir)
+            mkfs_ext4(img_file, contents_dir, 'snap')
             # Because there were no contents, the `sudo cp` was never called,
             # the mock's shutil.copytree() was also never called, therefore
             # the results_dir was never created.
@@ -394,7 +555,7 @@ class TestHelpers(TestCase):
                 fp.write(b'01234')
             # And a fake image file.
             img_file = resources.enter_context(NamedTemporaryFile())
-            mkfs_ext4(img_file, contents_dir, preserve_ownership=True)
+            mkfs_ext4(img_file, contents_dir, 'snap', preserve_ownership=True)
             with open(os.path.join(mock.results_dir, 'a.dat'), 'rb') as fp:
                 self.assertEqual(fp.read(), b'01234')
             self.assertTrue(mock.preserves_ownership)
@@ -419,3 +580,37 @@ class TestHelpers(TestCase):
             mock_geteuid.assert_called_with()
             self.assertEqual(
                 cm.exception.user_name, 'test')
+
+    def test_unsparse_swapfile_ext4(self):
+        with ExitStack() as resources:
+            tmpdir = resources.enter_context(TemporaryDirectory())
+            results_dir = os.path.join(tmpdir, 'results')
+            os.makedirs(results_dir)
+            # Empty swapfile.
+            swapfile = os.path.join(results_dir, 'swapfile')
+            run('dd if=/dev/zero of={} bs=10M count=1'.format(swapfile),
+                stdout=None, stderr=None, capture_output=False)
+            # Image file.
+            img_file = resources.enter_context(NamedTemporaryFile())
+            mock = MountMocker(results_dir, results_dir)
+            resources.enter_context(
+                patch('ubuntu_image.helpers.run', mock.run))
+            # Now the actual test.
+            unsparse_swapfile_ext4(img_file)
+            self.assertTrue(mock.dd_called)
+
+    def test_unsparse_swapfile_ext4_no_swap(self):
+        with ExitStack() as resources:
+            tmpdir = resources.enter_context(TemporaryDirectory())
+            results_dir = os.path.join(tmpdir, 'results')
+            os.makedirs(results_dir)
+            # No swapfile, just some file.
+            open(os.path.join(results_dir, 'foobar'), 'w')
+            # Image file.
+            img_file = resources.enter_context(NamedTemporaryFile())
+            mock = MountMocker(results_dir, results_dir)
+            resources.enter_context(
+                patch('ubuntu_image.helpers.run', mock.run))
+            # Now the actual test.
+            unsparse_swapfile_ext4(img_file)
+            self.assertFalse(mock.dd_called)
